import hashlib
from decimal import Decimal
from io import BytesIO
from math import ceil, floor, log
from typing import Tuple, List, Union
from stellaris.constants import MAX_SUPPLY, ENDIAN, MAX_BLOCK_SIZE_HEX
from stellaris.database import Database

BLOCK_TIME = 180
BLOCKS_COUNT = Decimal(500)
START_DIFFICULTY = Decimal('6.0')

def difficulty_to_hashrate_old(difficulty: Decimal) -> int:
    decimal = difficulty % 1 or 1/16
    return Decimal(16 ** int(difficulty) * (16 * decimal))


def difficulty_to_hashrate(difficulty: Decimal) -> int:
    decimal = difficulty % 1
    return Decimal(16 ** int(difficulty) * (16 / ceil(16 * (1 - decimal))))


def hashrate_to_difficulty_old(hashrate: int) -> Decimal:
    difficulty = int(log(hashrate, 16))
    if hashrate == 16 ** difficulty:
        return Decimal(difficulty)
    return Decimal(difficulty + (hashrate / Decimal(16) ** difficulty) / 16)


def hashrate_to_difficulty_wrong(hashrate: int) -> Decimal:
    difficulty = int(log(hashrate, 16))
    if hashrate == 16 ** difficulty:
        return Decimal(difficulty)
    ratio = hashrate / 16 ** difficulty

    decimal = 16 / ratio / 16
    decimal = 1 - floor(decimal * 10) / Decimal(10)
    return Decimal(difficulty + decimal)


def hashrate_to_difficulty(hashrate: int) -> Decimal:
    difficulty = int(log(hashrate, 16))
    ratio = hashrate / 16 ** difficulty

    for i in range(0, 10):
        coeff = 16 / ceil(16 * (1 - i / 10))
        if coeff > ratio:
            decimal = (i - 1) / Decimal(10)
            return Decimal(difficulty + decimal)
        if coeff == ratio:
            decimal = i / Decimal(10)
            return Decimal(difficulty + decimal)

    return Decimal(difficulty) + Decimal('0.9')


async def calculate_difficulty() -> Tuple[Decimal, dict]:
    database = Database.instance
    last_block = await database.get_last_block()
    if last_block is None:
        return START_DIFFICULTY, dict()
    last_block = dict(last_block)
    last_block['address'] = last_block['address'].strip(' ')
    if last_block['id'] < BLOCKS_COUNT:
        return START_DIFFICULTY, last_block

    if last_block['id'] % BLOCKS_COUNT == 0:
        last_adjust_block = await database.get_block_by_id(last_block['id'] - BLOCKS_COUNT + 1)
        elapsed = last_block['timestamp'] - last_adjust_block['timestamp']
        average_per_block = elapsed / BLOCKS_COUNT
        last_difficulty = last_block['difficulty']
        if last_block['id'] <= 17500:
            hashrate = difficulty_to_hashrate_old(last_difficulty)
        else:
            hashrate = difficulty_to_hashrate(last_difficulty)
        ratio = BLOCK_TIME / average_per_block
        if last_block['id'] >= 180_000:  # from block 180k, allow difficulty to double at most
            ratio = min(ratio, 2)
        hashrate *= ratio
        if last_block['id'] < 17500:
            new_difficulty = hashrate_to_difficulty_old(hashrate)
            new_difficulty = floor(new_difficulty * 10) / Decimal(10)
        elif last_block['id'] < 180_000:
            new_difficulty = hashrate_to_difficulty_wrong(hashrate)
        else:
            new_difficulty = hashrate_to_difficulty(hashrate)
        return new_difficulty, last_block

    return last_block['difficulty'], last_block


from decimal import Decimal

from stellaris.constants import ENDIAN
from stellaris.utils.general import sha256
from stellaris.transactions import TransactionOutput


class CoinbaseTransaction:
    _hex: str = None

    def __init__(self, block_hash: str, address: str, amount: Decimal):
        self.block_hash = block_hash
        self.address = address
        self.amount = amount
        self.outputs = [TransactionOutput(address, amount)]

    async def verify(self):
        from stellaris.database import Database
        block = await (await Database.get()).get_block(self.block_hash)
        return block['address'] == self.address and self.amount == block['reward']

    def hex(self):
        if self._hex is not None:
            return self._hex
        hex_inputs = (bytes.fromhex(self.block_hash) + (0).to_bytes(1, ENDIAN)).hex()
        hex_outputs = ''.join(tx_output.tobytes().hex() for tx_output in self.outputs)

        if all(len(tx_output.address_bytes) == 64 for tx_output in self.outputs):
            version = 1
        elif all(len(tx_output.address_bytes) == 33 for tx_output in self.outputs):
            version = 2
        else:
            raise NotImplementedError()

        self._hex = ''.join([
            version.to_bytes(1, ENDIAN).hex(),
            (1).to_bytes(1, ENDIAN).hex(),
            hex_inputs,
            (1).to_bytes(1, ENDIAN).hex(),
            hex_outputs,
            (36).to_bytes(1, ENDIAN).hex(),
        ])

        return self._hex

    def hash(self):
        return sha256(self.hex())


import json
import os
from os.path import dirname, exists
from random import sample

import httpx
import pickledb

from stellaris.constants import MAX_BLOCK_SIZE_HEX
from stellaris.utils.general import timestamp

ACTIVE_NODES_DELTA = 60 * 60 * 24 * 7  # 7 days
INACTIVE_NODES_DELTA = 60 * 60 * 24 * 90  # 3 months
MAX_NODES_COUNT = 100

path = dirname(os.path.realpath(__file__)) + '/nodes.json'
if not exists(path):
    json.dump({}, open(path, 'wt'))
db = pickledb.load(path, True)


class NodesManager:
    last_messages: dict = None
    nodes: list = None
    db = db

    timeout = httpx.Timeout(3)
    async_client = httpx.AsyncClient(timeout=timeout, follow_redirects=True)

    @staticmethod
    def init():
        NodesManager.db._loaddb()
        NodesManager.nodes = NodesManager.db.get('nodes') or ['https://stellaris-node.connor33341.dev']
        NodesManager.last_messages = NodesManager.db.get('last_messages') or {'https://stellaris-node.connor33341.dev': timestamp()}

    @staticmethod
    def sync():
        NodesManager.db.set('nodes', NodesManager.nodes)
        NodesManager.db.set('last_messages', NodesManager.last_messages)

    @staticmethod
    async def request(url: str, method: str = 'GET', **kwargs):
        async with NodesManager.async_client.stream(method, url, **kwargs) as response:
            res = ''
            async for chunk in response.aiter_text():
                res += chunk
                if len(res) > MAX_BLOCK_SIZE_HEX * 10:
                    break
        return json.loads(res)

    @staticmethod
    async def is_node_working(node: str):
        try:
            await NodesManager.request(node)
            return True
        except:
            return False

    @staticmethod
    def add_node(node: str):
        node = node.strip('/')
        if len(NodesManager.nodes) > MAX_NODES_COUNT or len(NodesManager.get_zero_nodes()) > 10:
            NodesManager.clear_old_nodes()
        if len(NodesManager.nodes) > MAX_NODES_COUNT:
            raise Exception('Too many nodes')
        NodesManager.init()
        NodesManager.nodes.append(node)
        NodesManager.sync()

    @staticmethod
    def get_nodes():
        NodesManager.init()
        NodesManager.nodes.extend(NodesManager.last_messages.keys())
        NodesManager.nodes = [node.strip('/') for node in NodesManager.nodes if len(node)]
        NodesManager.nodes = list(dict.fromkeys(NodesManager.nodes))
        NodesManager.sync()
        return NodesManager.nodes

    @staticmethod
    def get_recent_nodes():
        full_nodes = {node_url: NodesManager.get_last_message(node_url) for node_url in NodesManager.get_nodes()}
        return [item[0] for item in sorted(full_nodes.items(), key=lambda item: item[1], reverse=True) if item[1] > timestamp() - ACTIVE_NODES_DELTA]

    @staticmethod
    def get_zero_nodes():
        return [node for node in NodesManager.get_nodes() if NodesManager.get_last_message(node) == 0]

    @staticmethod
    def get_propagate_nodes():
        active_nodes = NodesManager.get_recent_nodes()
        zero_nodes = NodesManager.get_zero_nodes()
        return (sample(active_nodes, k=10) if len(active_nodes) > 10 else active_nodes) + (sample(zero_nodes, k=10) if len(zero_nodes) > 10 else zero_nodes)
    @staticmethod
    def clear_old_nodes():
        NodesManager.init()
        NodesManager.nodes = [node for node in NodesManager.get_nodes() if NodesManager.get_last_message(node) > timestamp() - INACTIVE_NODES_DELTA]
        NodesManager.sync()

    @staticmethod
    def get_last_message(node_url: str):
        NodesManager.init()
        last_messages = NodesManager.last_messages
        return last_messages[node_url] if node_url in last_messages else 0

    @staticmethod
    def update_last_message(node_url: str):
        NodesManager.init()
        NodesManager.last_messages[node_url.strip('/')] = timestamp()
        NodesManager.sync()


class NodeInterface:
    def __init__(self, url: str):
        self.url = url.strip('/')
        self.base_url = self.url.replace('http://', '', 1).replace('https://', '', 1)

    async def get_block(self, block_no: int, full_transactions: bool = False):
        res = await self.request('get_block', {'block': block_no, 'full_transactions': full_transactions})
        return res['result']

    async def get_blocks(self, offset: int, limit: int):
        res = await self.request('get_blocks', {'offset': offset, 'limit': limit})
        if 'result' not in res:
            # todo improve error handling
            raise Exception(res['error'])
        return res['result']

    async def get_nodes(self):
        res = await self.request('get_nodes')
        return res['result']

    async def request(self, path: str, data: dict = {}, sender_node: str = ''):
        headers = {'Sender-Node': sender_node}
        if path in ('push_block', 'push_tx'):
            r = await NodesManager.request(f'{self.url}/{path}', method='POST', json=data, headers=headers, timeout=10)
        else:
            r = await NodesManager.request(f'{self.url}/{path}', params=data, headers=headers, timeout=10)
        return r


from decimal import Decimal
from typing import Tuple

from fastecdsa import ecdsa
from fastecdsa.point import Point

from stellaris.constants import CURVE, ENDIAN, SMALLEST
from stellaris.utils.general import point_to_string, string_to_point


class TransactionInput:
    public_key = None

    signed: Tuple[int, int] = None
    amount: Decimal = None

    def __init__(self, input_tx_hash: str, index: int, private_key: int = None, transaction=None, amount: Decimal = None, public_key: Point = None):
        self.tx_hash = input_tx_hash
        self.index = index
        self.private_key = private_key
        self.transaction = transaction
        self.transaction_info = None
        self.amount = amount
        self.public_key = public_key
        if transaction is not None and amount is None:
            self.get_related_output()

    async def get_transaction(self):
        if self.transaction is None:
            from stellaris.database import Database
            self.transaction = await Database.instance.get_transaction(self.tx_hash, check_signatures=False)
            assert self.transaction is not None
        return self.transaction

    async def get_transaction_info(self):
        if self.transaction_info is None:
            from stellaris.database import Database
            self.transaction_info = await Database.instance.get_transaction_info(self.tx_hash)
        assert self.transaction_info is not None
        return self.transaction_info

    async def get_related_output(self):
        tx = await self.get_transaction()
        related_output = tx.outputs[self.index]
        self.amount = related_output.amount
        return related_output

    async def get_related_output_info(self):
        tx = await self.get_transaction_info()
        related_output = {'address': tx['outputs_addresses'][self.index], 'amount': Decimal(tx['outputs_amounts'][self.index]) / SMALLEST}
        self.amount = related_output['amount']
        return related_output

    async def get_amount(self):
        if self.amount is None:
            if self.transaction is not None:
                return self.transaction.outputs[self.index].amount
            else:
                await self.get_related_output_info()
        return self.amount

    async def get_address(self):
        if self.transaction is not None:
            return (await self.get_related_output()).address
        return (await self.get_related_output_info())['address']

    def sign(self, tx_hex: str, private_key: int = None):
        private_key = private_key if private_key is not None else self.private_key
        self.signed = ecdsa.sign(bytes.fromhex(tx_hex), private_key)

    async def get_public_key(self):
        return self.public_key or string_to_point(await self.get_address())

    def tobytes(self):
        return bytes.fromhex(self.tx_hash) + self.index.to_bytes(1, ENDIAN)

    def get_signature(self):
        return self.signed[0].to_bytes(32, ENDIAN).hex() + self.signed[1].to_bytes(32, ENDIAN).hex()

    async def verify(self, input_tx) -> bool:
        try:
            public_key = await self.get_public_key()
        except AssertionError:
            return False
        # print('verifying with', point_to_string(public_key))

        return \
            ecdsa.verify(self.signed, bytes.fromhex(input_tx), public_key, CURVE) or \
            ecdsa.verify(self.signed, input_tx, public_key, CURVE)

    @property
    def as_dict(self):
        self_dict = vars(self).copy()
        self_dict['signed'] = self_dict['signed'] is not None
        if 'public_key' in self_dict: self_dict['public_key'] = point_to_string(self_dict['public_key'])
        if 'transaction' in self_dict: del self_dict['transaction']
        if 'private_key' in self_dict: del self_dict['private_key']
        return self_dict

    def __eq__(self, other):
        assert isinstance(other, self.__class__)
        return (self.tx_hash, self.index) == (other.tx_hash, other.index)


from decimal import Decimal

from stellaris.constants import ENDIAN, SMALLEST, CURVE
from stellaris.utils.general import byte_length, string_to_point, string_to_bytes


class TransactionOutput:
    def __init__(self, address: str, amount: Decimal):
        from fastecdsa.point import Point
        if isinstance(address, Point):
            raise Exception('TransactionOutput does not accept Point anymore. Pass the address string instead')
        self.address = address
        self.address_bytes = string_to_bytes(address)
        self.public_key = string_to_point(address)
        assert (amount * SMALLEST) % 1 == 0.0, 'too many decimal digits'
        self.amount = amount

    def tobytes(self):
        amount = int(self.amount * SMALLEST)
        count = byte_length(amount)
        return self.address_bytes + count.to_bytes(1, ENDIAN) + amount.to_bytes(count, ENDIAN)

    def verify(self):
        return self.amount > 0 and CURVE.is_point_on_curve((self.public_key.x, self.public_key.y))

    @property
    def as_dict(self):
        res = vars(self).copy()
        if 'public_key' in res: del res['public_key']
        return res


import hashlib
from decimal import Decimal
from io import BytesIO
from math import ceil, floor, log
from typing import Tuple, List, Union

from stellaris.database import Database, OLD_BLOCKS_TRANSACTIONS_ORDER
from stellaris.constants import MAX_SUPPLY, ENDIAN, MAX_BLOCK_SIZE_HEX
from stellaris.utils.general import sha256, timestamp, bytes_to_string, string_to_bytes
from stellaris.transactions import CoinbaseTransaction, Transaction
from stellaris.utils.block_utils import calculate_difficulty, difficulty_to_hashrate, difficulty_to_hashrate_old, hashrate_to_difficulty, hashrate_to_difficulty_old, hashrate_to_difficulty_wrong, BLOCK_TIME, BLOCKS_COUNT, START_DIFFICULTY

async def get_difficulty() -> Tuple[Decimal, dict]:
    if Manager.difficulty is None:
        Manager.difficulty = await calculate_difficulty()
    return Manager.difficulty


async def check_block_is_valid(block_content: str, mining_info: tuple = None) -> bool:
    if mining_info is None:
        mining_info = await get_difficulty()
    difficulty, last_block = mining_info

    block_hash = sha256(block_content)

    if 'hash' not in last_block:
        return True

    last_block_hash = last_block['hash']

    decimal = difficulty % 1
    difficulty = floor(difficulty)
    if decimal > 0:
        charset = '0123456789abcdef'
        count = ceil(16 * (1 - decimal))
        return block_hash.startswith(last_block_hash[-difficulty:]) and block_hash[difficulty] in charset[:count]
    return block_hash.startswith(last_block_hash[-difficulty:])


def get_block_reward(number: int) -> Decimal:
    divider = floor(number / 150000)
    if divider == 0:
        return Decimal(100)
    if divider > 8:
        if number < 150000 * 9 + 458732 - 150000:
            return Decimal('0.390625')
        elif number < 150000 * 9 + 458733 - 150000 + 320:
            return Decimal('0.3125')
        return Decimal(0)
    return Decimal(100) / (2 ** Decimal(divider))


def __check():
    i = 1
    r = 0
    index = {}
    while n := get_block_reward(i):
        if n not in index:
            index[n] = 0
        index[n] += 1
        i += 1
        r += n

    print(r)
    print(MAX_SUPPLY - r)
    print(index)


async def clear_pending_transactions(transactions=None):
    database: Database = Database.instance
    await database.clear_duplicate_pending_transactions()
    transactions = transactions or await database.get_pending_transactions_limit(hex_only=True)
    used_inputs = []
    for transaction in transactions:
        if isinstance(transaction, str):
            tx_hash = sha256(transaction)
            transaction = await Transaction.from_hex(transaction, check_signatures=False)
        else:
            tx_hash = sha256(transaction.hex())
        tx_inputs = [(tx_input.tx_hash, tx_input.index) for tx_input in transaction.inputs]
        if any(used_input in tx_inputs for used_input in used_inputs):
            await database.remove_pending_transaction(tx_hash)
            print(f'removed {tx_hash}')
            return await clear_pending_transactions()
        used_inputs += tx_inputs
    unspent_outputs = await database.get_unspent_outputs(used_inputs)
    double_spend_inputs = set(used_inputs) - set(unspent_outputs)
    if double_spend_inputs == set(used_inputs):
        await database.remove_pending_transactions()
    elif double_spend_inputs:
        await database.remove_pending_transactions_by_contains([tx_input[0] + bytes([tx_input[1]]).hex() for tx_input in double_spend_inputs])


def get_transactions_merkle_tree_ordered(transactions: List[Union[Transaction, str]]):
    _bytes = bytes()
    for transaction in transactions:
        _bytes += hashlib.sha256(bytes.fromhex(transaction.hex() if isinstance(transaction, Transaction) else transaction)).digest()
    return hashlib.sha256(_bytes).hexdigest()


def get_transactions_merkle_tree(transactions: List[Union[Transaction, str]]):
    _bytes = bytes()
    transactions_bytes = []
    for transaction in transactions:
        transactions_bytes.append(bytes.fromhex(transaction.hex() if isinstance(transaction, Transaction) else transaction))
    for transaction in sorted(transactions_bytes):
        _bytes += hashlib.sha256(transaction).digest()
    return hashlib.sha256(_bytes).hexdigest()


def get_transactions_size(transactions: List[Transaction]):
    return sum(len(transaction.hex()) for transaction in transactions)


def block_to_bytes(last_block_hash: str, block: dict) -> bytes:
    address_bytes = string_to_bytes(block['address'])
    version = bytes([])
    if len(address_bytes) != 64:
        version = bytes([2])
    return version + \
           bytes.fromhex(last_block_hash) + \
           address_bytes + \
           bytes.fromhex(block['merkle_tree']) + \
           block['timestamp'].to_bytes(4, byteorder=ENDIAN) + \
           int(float(block['difficulty']) * 10).to_bytes(2, ENDIAN) \
           + block['random'].to_bytes(4, ENDIAN)


def split_block_content(block_content: str):
    _bytes = bytes.fromhex(block_content)
    stream = BytesIO(_bytes)
    if len(_bytes) == 138:
        version = 1
    else:
        version = int.from_bytes(stream.read(1), ENDIAN)
        assert version > 1
        if version == 2:
            assert len(_bytes) == 108
        else:
            raise NotImplementedError()
    previous_hash = stream.read(32).hex()
    address = bytes_to_string(stream.read(64 if version == 1 else 33))
    merkle_tree = stream.read(32).hex()
    timestamp = int.from_bytes(stream.read(4), ENDIAN)
    difficulty = int.from_bytes(stream.read(2), ENDIAN) / Decimal(10)
    random = int.from_bytes(stream.read(4), ENDIAN)
    return previous_hash, address, merkle_tree, timestamp, difficulty, random


async def check_block(block_content: str, transactions: List[Transaction], mining_info: tuple = None):
    if mining_info is None:
        mining_info = await calculate_difficulty()
    difficulty, last_block = mining_info
    block_no = last_block['id'] + 1 if last_block != {} else 1
    previous_hash, address, merkle_tree, content_time, content_difficulty, random = split_block_content(block_content)
    if block_no == 17972 and last_block['hash'] == 'c3b69440e58e99567571e58486d8f22ed1e3107c50b827c9366294b2637cb1a0':
        if address != 'dbda85e237b90aa669da00f2859e0010b0a62e0fb6e55ba6ca3ce8a961a60c64410bcfb6a038310a3bb6f1a4aaa2de1192cc10e380a774bb6f9c6ca8547f11ab' or \
           content_time != 1638463765 or random != 17660081:
            return False
    elif not await check_block_is_valid(block_content, mining_info):
        print('block not valid')
        return False
    if block_no == 143361 and sha256(block_content) == 'a53268dd22d173dd0c9c10d7f6a64f46071c669052186a7855e9cc65e9a46939':
        for transaction in transactions:
            if transaction.hash() == '5958b48fa0b1692b112affc7a2be887d24073027f3bef585322f33b5eeca463c':
                transactions.remove(transaction)  # there are 2 transactions which spend same inputs in this block
                break

    content_time = int(content_time)
    if last_block != {} and previous_hash != last_block['hash']:
        return False

    if (last_block['timestamp'] if 'timestamp' in last_block else 0) > content_time:
        print('timestamp younger than previous block')
        return False

    if block_no >= 291500 and (last_block['timestamp'] if 'timestamp' in last_block else 0) == content_time:
        print('timestamp younger than previous block')
        return False

    if content_time > timestamp():
        print('timestamp in the future')
        return False

    database: Database = Database.instance
    transactions = [tx for tx in transactions if isinstance(tx, Transaction)]
    if get_transactions_size(transactions) > MAX_BLOCK_SIZE_HEX:
        print('block is too big')
        return False

    if transactions:
        check_inputs = sum([[(tx_input.tx_hash, tx_input.index) for tx_input in transaction.inputs] for transaction in transactions], [])
        unspent_outputs = await database.get_unspent_outputs(check_inputs)
        if len(set(check_inputs)) != len(check_inputs) or set(check_inputs) - set(unspent_outputs) != set():
            print('double spend in block')
            spent_outputs = set(check_inputs) - set(unspent_outputs)
            print(len(spent_outputs))
            return False
        input_txs_hash = sum([[tx_input.tx_hash for tx_input in transaction.inputs] for transaction in transactions], [])
        input_txs = await database.get_transactions_info(input_txs_hash)
        # move after pp('after get_transactions', time.time() - t)
        for transaction in transactions:
            await transaction._fill_transaction_inputs(input_txs)

    for transaction in transactions:
        if not await transaction.verify(check_double_spend=False):
            print(f'transaction {transaction.hash()} has been not verified')
            return False

    transactions_merkle_tree = get_transactions_merkle_tree(
        transactions) if block_no >= 22500 else get_transactions_merkle_tree_ordered(transactions)
    if merkle_tree != transactions_merkle_tree:
        if block_no == 17972 and get_transactions_merkle_tree(transactions) == 'cb52390983d1902bf7d0eb96ed3f8adc359d34b6617dcccd2b610349e0ee8d15':
            return True
        if block_no == 143361 and transactions_merkle_tree == 'a9a930d5144c70afc1679dbb83551a318d5d5da6744145761962157a48fabd54':
            return True
        print('merkle tree does not match')
        return False

    return True


async def create_block(block_content: str, transactions: List[Transaction], last_block: dict = None):
    Manager.difficulty = None
    if last_block is None or last_block['id'] % BLOCKS_COUNT == 0:
        difficulty, last_block = await calculate_difficulty()
    else:
        # fixme temp fix
        difficulty, last_block = await get_difficulty()
        #difficulty = Decimal(str(last_block['difficulty']))
    if not await check_block(block_content, transactions, (difficulty, last_block)):
        return False

    database: Database = Database.instance
    block_no = last_block['id'] + 1 if last_block != {} else 1
    block_hash = sha256(block_content) if block_no != 17972 else '37cb1a0522c039330775e07d824c94e0422dbfb2dba6dcd421f4dc9f11601672'
    previous_hash, address, merkle_tree, content_time, content_difficulty, random = split_block_content(block_content)
    if block_hash == 'a53268dd22d173dd0c9c10d7f6a64f46071c669052186a7855e9cc65e9a46939':  # block 143361 has a double spend
        for transaction in transactions:
            if transaction.hash() == '5958b48fa0b1692b112affc7a2be887d24073027f3bef585322f33b5eeca463c':
                transactions.remove(transaction)  # there are 2 transactions which spend same inputs in this block
                break
    fees = sum(transaction.fees for transaction in transactions)

    block_reward = get_block_reward(block_no)
    coinbase_transaction = CoinbaseTransaction(block_hash, address, block_reward + fees)
    if block_no > 35000:
        if not coinbase_transaction.outputs[0].verify():
            return False

    await database.add_block(block_no, block_hash, block_content, address, random, difficulty, block_reward + fees, content_time)
    await database.add_transaction(coinbase_transaction, block_hash)

    try:
        await database.add_transactions(transactions, block_hash)
        if len(transactions) > 1 and block_no < 22500:
            OLD_BLOCKS_TRANSACTIONS_ORDER.set(block_hash, [transaction.hex() for transaction in transactions])
    except Exception as e:
        print(f'a transaction has not been added in block', e)
        await database.delete_block(block_no)
        return False
    await database.add_unspent_transactions_outputs(transactions + [coinbase_transaction])
    if transactions:
        await database.remove_pending_transactions_by_hash([transaction.hash() for transaction in transactions])
        await database.remove_unspent_outputs(transactions)
        await database.remove_pending_spent_outputs(transactions)

        print(f'Added {len(transactions)} transactions in block {block_no}. Reward: {block_reward}, Fees: {fees}')
    Manager.difficulty = None
    return True


class Manager:
    difficulty: Tuple[float, dict] = None


import os
import json
import gzip
import asyncio
from datetime import datetime, timezone
from decimal import Decimal
from statistics import mean
from typing import List, Union, Tuple, Dict
from pathlib import Path

import pickledb

from stellaris.constants import MAX_BLOCK_SIZE_HEX, SMALLEST
from stellaris.utils.general import sha256, point_to_string, string_to_point, point_to_bytes, AddressFormat, normalize_block
from stellaris.transactions import Transaction, CoinbaseTransaction, TransactionInput

dir_path = os.path.dirname(os.path.realpath(__file__))
OLD_BLOCKS_TRANSACTIONS_ORDER = pickledb.load(dir_path + '/old_block_transactions_order.json', True)


class Database:
    instance = None
    
    def __init__(self):
        self.data_dir = None
        self.blocks_file = None
        self.transactions_file = None
        self.pending_transactions_file = None
        self.unspent_outputs_file = None
        self.pending_spent_outputs_file = None
        self._blocks = {}
        self._transactions = {}
        self._pending_transactions = {}
        self._unspent_outputs = set()
        self._pending_spent_outputs = set()
        self._transaction_block_map = {}
        self.is_indexed = True
        self._lock = asyncio.Lock()

    @staticmethod
    async def create(data_dir='./data/database', **kwargs):
        self = Database()
        self.data_dir = Path(data_dir)
        self.data_dir.mkdir(parents=True, exist_ok=True)
        
        self.blocks_file = self.data_dir / 'blocks.json.gz'
        self.transactions_file = self.data_dir / 'transactions.json.gz'
        self.pending_transactions_file = self.data_dir / 'pending_transactions.json.gz'
        self.unspent_outputs_file = self.data_dir / 'unspent_outputs.json.gz'
        self.pending_spent_outputs_file = self.data_dir / 'pending_spent_outputs.json.gz'
        
        await self._load_data()
        Database.instance = self
        return self

    @staticmethod
    async def get():
        if Database.instance is None:
            await Database.create()
        return Database.instance

    async def _save_to_file(self, file_path: Path, data):
        """Save data to compressed JSON file"""
        async with self._lock:
            with gzip.open(file_path, 'wt', encoding='utf-8') as f:
                json.dump(data, f, indent=2, default=str)

    async def _load_from_file(self, file_path: Path):
        """Load data from compressed JSON file"""
        if not file_path.exists():
            return {}
        try:
            with gzip.open(file_path, 'rt', encoding='utf-8') as f:
                return json.load(f)
        except (json.JSONDecodeError, OSError):
            return {}

    async def _load_data(self):
        """Load all data from files"""
        self._blocks = await self._load_from_file(self.blocks_file)
        self._transactions = await self._load_from_file(self.transactions_file)
        self._pending_transactions = await self._load_from_file(self.pending_transactions_file)
        
        unspent_data = await self._load_from_file(self.unspent_outputs_file)
        self._unspent_outputs = set(tuple(item) for item in unspent_data.get('outputs', []))
        
        pending_spent_data = await self._load_from_file(self.pending_spent_outputs_file)
        self._pending_spent_outputs = set(tuple(item) for item in pending_spent_data.get('outputs', []))
        
        # Build transaction to block mapping
        for tx_hash, tx_data in self._transactions.items():
            if 'block_hash' in tx_data:
                self._transaction_block_map[tx_hash] = tx_data['block_hash']

    async def _save_blocks(self):
        await self._save_to_file(self.blocks_file, self._blocks)

    async def _save_transactions(self):
        await self._save_to_file(self.transactions_file, self._transactions)

    async def _save_pending_transactions(self):
        await self._save_to_file(self.pending_transactions_file, self._pending_transactions)

    async def _save_unspent_outputs(self):
        data = {'outputs': list(self._unspent_outputs)}
        await self._save_to_file(self.unspent_outputs_file, data)

    async def _save_pending_spent_outputs(self):
        data = {'outputs': list(self._pending_spent_outputs)}
        await self._save_to_file(self.pending_spent_outputs_file, data)

    async def add_pending_transaction(self, transaction: Transaction, verify: bool = True):
        if isinstance(transaction, CoinbaseTransaction):
            return False
        tx_hex = transaction.hex()
        if verify and not await transaction.verify_pending():
            return False
        
        tx_hash = sha256(tx_hex)
        utc_datetime = datetime.now(timezone.utc)
        
        self._pending_transactions[tx_hash] = {
            'tx_hash': tx_hash,
            'tx_hex': tx_hex,
            'inputs_addresses': [point_to_string(await tx_input.get_public_key()) for tx_input in transaction.inputs],
            'fees': transaction.fees,
            'time_received': utc_datetime.isoformat(),
            'propagation_time': utc_datetime.isoformat()
        }
        
        await self._save_pending_transactions()
        await self.add_transactions_pending_spent_outputs([transaction])
        return True

    async def remove_pending_transaction(self, tx_hash: str):
        if tx_hash in self._pending_transactions:
            del self._pending_transactions[tx_hash]
            await self._save_pending_transactions()

    async def remove_pending_transactions_by_hash(self, tx_hashes: List[str]):
        for tx_hash in tx_hashes:
            if tx_hash in self._pending_transactions:
                del self._pending_transactions[tx_hash]
        await self._save_pending_transactions()

    async def remove_pending_transactions(self):
        self._pending_transactions.clear()
        await self._save_pending_transactions()

    async def delete_blockchain(self):
        self._blocks.clear()
        self._transactions.clear()
        self._transaction_block_map.clear()
        await self._save_blocks()
        await self._save_transactions()

    async def delete_block(self, id: int):
        block_to_remove = None
        for block_hash, block_data in self._blocks.items():
            if block_data.get('id') == id:
                block_to_remove = block_hash
                break
        
        if block_to_remove:
            # Remove transactions for this block
            txs_to_remove = []
            for tx_hash, tx_data in self._transactions.items():
                if tx_data.get('block_hash') == block_to_remove:
                    txs_to_remove.append(tx_hash)
            
            for tx_hash in txs_to_remove:
                del self._transactions[tx_hash]
                if tx_hash in self._transaction_block_map:
                    del self._transaction_block_map[tx_hash]
            
            del self._blocks[block_to_remove]
            await self._save_blocks()
            await self._save_transactions()

    async def delete_blocks(self, offset: int):
        blocks_to_remove = []
        for block_hash, block_data in self._blocks.items():
            if block_data.get('id', 0) > offset:
                blocks_to_remove.append(block_hash)
        
        for block_hash in blocks_to_remove:
            await self.delete_block(self._blocks[block_hash]['id'])

    async def remove_blocks(self, block_no: int):
        blocks_to_remove = await self.get_blocks(block_no, 500)
        transactions_to_remove = []
        transactions_hashes = []
        
        for block_data in blocks_to_remove:
            block = block_data['block']
            transactions = block_data['transactions']
            transactions_to_remove.extend([await Transaction.from_hex(tx, False) for tx in transactions])
            transactions_hashes.extend([sha256(tx) for tx in transactions])
        
        outputs_to_be_restored = []
        for transaction in transactions_to_remove:
            if isinstance(transaction, Transaction):
                outputs_to_be_restored.extend([(tx_input.tx_hash, tx_input.index) for tx_input in transaction.inputs if tx_input.tx_hash not in transactions_hashes])
        
        # Remove blocks and their transactions
        for block_data in blocks_to_remove:
            block = block_data['block']
            block_hash = block['hash']
            if block_hash in self._blocks:
                del self._blocks[block_hash]
                
            # Remove transactions for this block
            txs_to_remove = []
            for tx_hash, tx_data in self._transactions.items():
                if tx_data.get('block_hash') == block_hash:
                    txs_to_remove.append(tx_hash)
            
            for tx_hash in txs_to_remove:
                del self._transactions[tx_hash]
                if tx_hash in self._transaction_block_map:
                    del self._transaction_block_map[tx_hash]
        
        await self._save_blocks()
        await self._save_transactions()
        await self.add_unspent_outputs(outputs_to_be_restored)

    async def get_pending_transactions_limit(self, limit: int = MAX_BLOCK_SIZE_HEX, hex_only: bool = False, check_signatures: bool = True) -> List[Union[Transaction, str]]:
        # Sort by fee efficiency (fees per byte), then by size, then by tx_hex
        pending_txs = list(self._pending_transactions.values())
        pending_txs.sort(key=lambda tx: (-tx['fees'] / len(tx['tx_hex']), len(tx['tx_hex']), tx['tx_hex']))
        
        return_txs = []
        size = 0
        for tx_data in pending_txs:
            tx_hex = tx_data['tx_hex']
            if size + len(tx_hex) > limit:
                break
            return_txs.append(tx_hex)
            size += len(tx_hex)
        
        if hex_only:
            return return_txs
        return [await Transaction.from_hex(tx_hex, check_signatures) for tx_hex in return_txs]

    async def get_need_propagate_transactions(self, last_propagation_delta: int = 600, limit: int = MAX_BLOCK_SIZE_HEX) -> List[Union[Transaction, str]]:
        current_time = datetime.now(timezone.utc)
        pending_txs = list(self._pending_transactions.values())
        pending_txs.sort(key=lambda tx: (-tx['fees'] / len(tx['tx_hex']), len(tx['tx_hex']), tx['tx_hex']))
        
        return_txs = []
        size = 0
        for tx_data in pending_txs:
            tx_hex = tx_data['tx_hex']
            if size + len(tx_hex) > limit:
                break
            size += len(tx_hex)
            
            propagation_time = datetime.fromisoformat(tx_data['propagation_time'].replace('Z', '+00:00'))
            if (current_time - propagation_time).total_seconds() > last_propagation_delta:
                return_txs.append(tx_hex)
        
        return return_txs

    async def update_pending_transactions_propagation_time(self, txs_hash: List[str]):
        current_time = datetime.now(timezone.utc)
        for tx_hash in txs_hash:
            if tx_hash in self._pending_transactions:
                self._pending_transactions[tx_hash]['propagation_time'] = current_time.isoformat()
        await self._save_pending_transactions()

    async def get_next_block_average_fee(self):
        limit = MAX_BLOCK_SIZE_HEX
        pending_txs = list(self._pending_transactions.values())
        pending_txs.sort(key=lambda tx: (-tx['fees'] / len(tx['tx_hex']), len(tx['tx_hex'])))
        
        fees = []
        size = 0
        for tx_data in pending_txs:
            tx_size = len(tx_data['tx_hex'])
            if size + tx_size > limit:
                break
            fees.append(tx_data['fees'])
            size += tx_size
        
        if not fees:
            return 0
        return int(mean(fees) * SMALLEST) // Decimal(SMALLEST)

    async def get_pending_blocks_count(self):
        total_size = sum(len(tx_data['tx_hex']) for tx_data in self._pending_transactions.values())
        return int(total_size / MAX_BLOCK_SIZE_HEX + 1)

    async def clear_duplicate_pending_transactions(self):
        to_remove = []
        for tx_hash in self._pending_transactions:
            if tx_hash in self._transactions:
                to_remove.append(tx_hash)
        
        for tx_hash in to_remove:
            del self._pending_transactions[tx_hash]
        
        if to_remove:
            await self._save_pending_transactions()

    async def add_transaction(self, transaction: Union[Transaction, CoinbaseTransaction], block_hash: str):
        await self.add_transactions([transaction], block_hash)

    async def add_transactions(self, transactions: List[Union[Transaction, CoinbaseTransaction]], block_hash: str):
        block_timestamp = None
        if block_hash in self._blocks:
            block_timestamp = self._blocks[block_hash].get('timestamp')
        
        for transaction in transactions:
            tx_hash = transaction.hash()
            
            # Get time_received from pending transactions or use block timestamp
            time_received = None
            if isinstance(transaction, Transaction) and tx_hash in self._pending_transactions:
                time_received = self._pending_transactions[tx_hash].get('time_received')
            elif isinstance(transaction, CoinbaseTransaction) and block_timestamp:
                time_received = block_timestamp
            
            self._transactions[tx_hash] = {
                'block_hash': block_hash,
                'tx_hash': tx_hash,
                'tx_hex': transaction.hex(),
                'inputs_addresses': [point_to_string(await tx_input.get_public_key()) for tx_input in transaction.inputs] if isinstance(transaction, Transaction) else [],
                'outputs_addresses': [tx_output.address for tx_output in transaction.outputs],
                'outputs_amounts': [tx_output.amount * SMALLEST for tx_output in transaction.outputs],
                'fees': transaction.fees if isinstance(transaction, Transaction) else 0,
                'time_received': time_received
            }
            
            self._transaction_block_map[tx_hash] = block_hash
        
        await self._save_transactions()

    async def add_block(self, id: int, block_hash: str, block_content: str, address: str, random: int, difficulty: Decimal, reward: Decimal, timestamp: Union[datetime, int]):
        if isinstance(timestamp, int):
            timestamp = datetime.utcfromtimestamp(timestamp)
        
        self._blocks[block_hash] = {
            'id': id,
            'hash': block_hash,
            'content': block_content,
            'address': address,
            'random': random,
            'difficulty': float(difficulty),
            'reward': float(reward),
            'timestamp': timestamp.isoformat()
        }
        
        await self._save_blocks()
        
        # Clear difficulty cache
        from stellaris.manager import Manager
        Manager.difficulty = None

    async def get_transaction(self, tx_hash: str, check_signatures: bool = True) -> Union[Transaction, CoinbaseTransaction]:
        if tx_hash not in self._transactions:
            return None
        
        tx_data = self._transactions[tx_hash]
        tx = await Transaction.from_hex(tx_data['tx_hex'], check_signatures)
        tx.block_hash = tx_data['block_hash']
        return tx

    async def get_transaction_info(self, tx_hash: str) -> dict:
        return self._transactions.get(tx_hash)

    async def get_transactions_info(self, tx_hashes: List[str]) -> Dict[str, dict]:
        return {tx_hash: self._transactions[tx_hash] for tx_hash in tx_hashes if tx_hash in self._transactions}

    async def get_pending_transaction(self, tx_hash: str, check_signatures: bool = True) -> Transaction:
        if tx_hash not in self._pending_transactions:
            return None
        
        tx_data = self._pending_transactions[tx_hash]
        return await Transaction.from_hex(tx_data['tx_hex'], check_signatures)

    async def get_pending_transactions_by_hash(self, hashes: List[str], check_signatures: bool = True) -> List[Transaction]:
        result = []
        for tx_hash in hashes:
            if tx_hash in self._pending_transactions:
                tx_data = self._pending_transactions[tx_hash]
                result.append(await Transaction.from_hex(tx_data['tx_hex'], check_signatures))
        return result

    async def get_transactions(self, tx_hashes: List[str]):
        result = {}
        for tx_hash in tx_hashes:
            if tx_hash in self._transactions:
                tx_data = self._transactions[tx_hash]
                tx = await Transaction.from_hex(tx_data['tx_hex'])
                result[sha256(tx_data['tx_hex'])] = tx
        return result

    async def get_transaction_hash_by_contains_multi(self, contains: List[str], ignore: str = None):
        for tx_hash, tx_data in self._transactions.items():
            if ignore and tx_hash == ignore:
                continue
            if any(contain in tx_data['tx_hex'] for contain in contains):
                return tx_hash
        return None

    async def get_pending_transactions_by_contains(self, contains: str):
        result = []
        for tx_hash, tx_data in self._pending_transactions.items():
            if contains in tx_data['tx_hex'] and tx_hash != contains:
                result.append(await Transaction.from_hex(tx_data['tx_hex']))
        return result if result else None

    async def remove_pending_transactions_by_contains(self, search: List[str]) -> None:
        to_remove = []
        for tx_hash, tx_data in self._pending_transactions.items():
            if any(s in tx_data['tx_hex'] for s in search):
                to_remove.append(tx_hash)
        
        for tx_hash in to_remove:
            del self._pending_transactions[tx_hash]
        
        if to_remove:
            await self._save_pending_transactions()

    async def get_pending_transaction_by_contains_multi(self, contains: List[str], ignore: str = None):
        for tx_hash, tx_data in self._pending_transactions.items():
            if ignore and tx_hash == ignore:
                continue
            if any(contain in tx_data['tx_hex'] for contain in contains):
                return await Transaction.from_hex(tx_data['tx_hex'])
        return None

    async def get_last_block(self) -> dict:
        if not self._blocks:
            return None
        
        # Find block with highest ID
        last_block = max(self._blocks.values(), key=lambda b: b.get('id', 0))
        return normalize_block(last_block)

    async def get_next_block_id(self) -> int:
        if not self._blocks:
            return 1
        
        last_id = max(block_data.get('id', 0) for block_data in self._blocks.values())
        return last_id + 1

    async def get_block(self, block_hash: str) -> dict:
        if block_hash not in self._blocks:
            return None
        
        return normalize_block(self._blocks[block_hash])

    async def get_blocks(self, offset: int, limit: int) -> list:
        # Get blocks in ID range
        selected_blocks = []
        for block_hash, block_data in self._blocks.items():
            block_id = block_data.get('id', 0)
            if block_id >= offset:
                selected_blocks.append((block_hash, block_data))
        
        # Sort by ID and limit
        selected_blocks.sort(key=lambda x: x[1].get('id', 0))
        selected_blocks = selected_blocks[:limit]
        
        # Get transactions for these blocks
        result = []
        total_size = 0
        
        for block_hash, block_data in selected_blocks:
            # Find transactions for this block
            block_transactions = []
            for tx_hash, tx_data in self._transactions.items():
                if tx_data.get('block_hash') == block_hash:
                    block_transactions.append(tx_data['tx_hex'])
            
            # Check if we have old transaction order data
            old_txs = OLD_BLOCKS_TRANSACTIONS_ORDER.get(block_hash)
            if old_txs:
                block_transactions = old_txs
            
            # Check size limit
            block_size = sum(len(tx) for tx in block_transactions)
            if total_size + block_size > MAX_BLOCK_SIZE_HEX * 8:
                break
            
            total_size += block_size
            result.append({
                'block': normalize_block(block_data),
                'transactions': block_transactions
            })
        
        return result

    async def get_block_by_id(self, block_id: int) -> dict:
        for block_data in self._blocks.values():
            if block_data.get('id') == block_id:
                return normalize_block(block_data)
        return None

    async def get_block_transactions(self, block_hash: str, check_signatures: bool = True, hex_only: bool = False) -> List[Union[Transaction, CoinbaseTransaction]]:
        transactions = []
        for tx_hash, tx_data in self._transactions.items():
            if tx_data.get('block_hash') == block_hash:
                if hex_only:
                    transactions.append(tx_data['tx_hex'])
                else:
                    transactions.append(await Transaction.from_hex(tx_data['tx_hex'], check_signatures))
        return transactions

    async def get_block_transaction_hashes(self, block_hash: str) -> List[str]:
        hashes = []
        for tx_hash, tx_data in self._transactions.items():
            if tx_data.get('block_hash') == block_hash and block_hash not in tx_data['tx_hex']:
                hashes.append(tx_hash)
        return hashes

    async def get_block_nice_transactions(self, block_hash: str) -> List[dict]:
        transactions = []
        for tx_hash, tx_data in self._transactions.items():
            if tx_data.get('block_hash') == block_hash:
                transactions.append({
                    'hash': tx_hash,
                    'is_coinbase': not tx_data.get('inputs_addresses', [])
                })
        return transactions

    async def add_unspent_outputs(self, outputs: List[Tuple[str, int]]) -> None:
        if not outputs:
            return
        
        for output in outputs:
            if len(output) == 2:
                self._unspent_outputs.add(output)
            elif len(output) == 3:
                # For outputs with address, we store as (tx_hash, index, address) but keep in set as (tx_hash, index)
                self._unspent_outputs.add((output[0], output[1]))
        
        await self._save_unspent_outputs()

    async def add_pending_spent_outputs(self, outputs: List[Tuple[str, int]]) -> None:
        for output in outputs:
            self._pending_spent_outputs.add(output)
        await self._save_pending_spent_outputs()

    async def add_transactions_pending_spent_outputs(self, transactions: List[Transaction]) -> None:
        outputs = []
        for transaction in transactions:
            outputs.extend([(tx_input.tx_hash, tx_input.index) for tx_input in transaction.inputs])
        
        for output in outputs:
            self._pending_spent_outputs.add(output)
        
        await self._save_pending_spent_outputs()

    async def add_unspent_transactions_outputs(self, transactions: List[Transaction]) -> None:
        outputs = []
        for transaction in transactions:
            tx_hash = transaction.hash()
            for index, output in enumerate(transaction.outputs):
                outputs.append((tx_hash, index, output.address))
        
        await self.add_unspent_outputs(outputs)

    async def remove_unspent_outputs(self, transactions: List[Transaction]) -> None:
        inputs = []
        for transaction in transactions:
            inputs.extend([(tx_input.tx_hash, tx_input.index) for tx_input in transaction.inputs])
        
        for input_tuple in inputs:
            self._unspent_outputs.discard(input_tuple)
        
        await self._save_unspent_outputs()

    async def remove_pending_spent_outputs(self, transactions: List[Transaction]) -> None:
        inputs = []
        for transaction in transactions:
            inputs.extend([(tx_input.tx_hash, tx_input.index) for tx_input in transaction.inputs])
        
        for input_tuple in inputs:
            self._pending_spent_outputs.discard(input_tuple)
        
        await self._save_pending_spent_outputs()

    async def get_unspent_outputs(self, outputs: List[Tuple[str, int]]) -> List[Tuple[str, int]]:
        result = []
        for output in outputs:
            if output in self._unspent_outputs:
                result.append(output)
        return result

    async def get_unspent_outputs_hash(self) -> str:
        # Sort outputs by tx_hash, index for consistent hashing
        sorted_outputs = sorted(self._unspent_outputs)
        hash_input = ''.join(tx_hash + bytes([index]).hex() for tx_hash, index in sorted_outputs)
        return sha256(hash_input)

    async def get_pending_spent_outputs(self, outputs: List[Tuple[str, int]]) -> List[Tuple[str, int]]:
        result = []
        for output in outputs:
            if output in self._pending_spent_outputs:
                result.append(output)
        return result

    async def set_unspent_outputs_addresses(self):
        # This method was used to populate address field in SQL - not needed for JSON storage
        # as we can store addresses directly when adding outputs
        pass

    async def get_unspent_outputs_from_all_transactions(self):
        # Rebuild unspent outputs from all transactions
        outputs = set()
        
        # Sort transactions by block ID
        sorted_transactions = []
        for tx_hash, tx_data in self._transactions.items():
            block_hash = tx_data.get('block_hash')
            if block_hash and block_hash in self._blocks:
                block_id = self._blocks[block_hash].get('id', 0)
                sorted_transactions.append((block_id, tx_hash, tx_data))
        
        sorted_transactions.sort(key=lambda x: x[0])  # Sort by block ID
        
        last_block_no = 0
        for block_id, tx_hash, tx_data in sorted_transactions:
            if block_id != last_block_no:
                last_block_no = block_id
                print(f'{len(outputs)} utxos at block {last_block_no - 1}')
            
            transaction = await Transaction.from_hex(tx_data['tx_hex'], check_signatures=False)
            
            # Remove spent outputs
            if isinstance(transaction, Transaction):
                spent_outputs = {(tx_input.tx_hash, tx_input.index) for tx_input in transaction.inputs}
                outputs = outputs.difference(spent_outputs)
            
            # Add new outputs
            new_outputs = {(tx_hash, index) for index in range(len(transaction.outputs))}
            outputs.update(new_outputs)
        
        return list(outputs)

    async def get_address_transactions(self, address: str, check_pending_txs: bool = False, check_signatures: bool = False, limit: int = 50, offset: int = 0) -> List[Union[Transaction, CoinbaseTransaction]]:
        point = string_to_point(address)
        search_patterns = [point_to_bytes(string_to_point(address), address_format).hex() for address_format in list(AddressFormat)]
        addresses = [point_to_string(point, address_format) for address_format in list(AddressFormat)]
        
        # Find transactions involving this address
        matching_txs = []
        
        # Search confirmed transactions
        for tx_hash, tx_data in self._transactions.items():
            block_hash = tx_data.get('block_hash')
            if block_hash in self._blocks:
                block_id = self._blocks[block_hash].get('id', 0)
                
                # Check if address is in inputs or outputs
                inputs_addresses = tx_data.get('inputs_addresses', [])
                outputs_addresses = tx_data.get('outputs_addresses', [])
                
                if (any(addr in addresses for addr in inputs_addresses) or 
                    any(addr in addresses for addr in outputs_addresses) or
                    any(pattern in tx_data['tx_hex'] for pattern in search_patterns)):
                    matching_txs.append((block_id, tx_data['tx_hex']))
        
        # Sort by block number descending
        matching_txs.sort(key=lambda x: x[0], reverse=True)
        
        # Apply pagination
        paginated_txs = matching_txs[offset:offset + limit]
        
        # Add pending transactions if requested
        if check_pending_txs:
            for tx_hash, tx_data in self._pending_transactions.items():
                inputs_addresses = tx_data.get('inputs_addresses', [])
                if (any(addr in addresses for addr in inputs_addresses) or
                    any(pattern in tx_data['tx_hex'] for pattern in search_patterns)):
                    paginated_txs.append((float('inf'), tx_data['tx_hex']))  # Pending txs have highest priority
        
        # Convert to Transaction objects
        result = []
        for _, tx_hex in paginated_txs:
            result.append(await Transaction.from_hex(tx_hex, check_signatures))
        
        return result

    async def get_address_pending_transactions(self, address: str, check_signatures: bool = False) -> List[Union[Transaction, CoinbaseTransaction]]:
        point = string_to_point(address)
        search_patterns = [point_to_bytes(string_to_point(address), address_format).hex() for address_format in list(AddressFormat)]
        addresses = [point_to_string(point, address_format) for address_format in list(AddressFormat)]
        
        result = []
        for tx_hash, tx_data in self._pending_transactions.items():
            inputs_addresses = tx_data.get('inputs_addresses', [])
            if (any(addr in addresses for addr in inputs_addresses) or
                any(pattern in tx_data['tx_hex'] for pattern in search_patterns)):
                result.append(await Transaction.from_hex(tx_data['tx_hex'], check_signatures))
        
        return result

    async def get_address_pending_spent_outputs(self, address: str, check_signatures: bool = False) -> List[Union[Transaction, CoinbaseTransaction]]:
        point = string_to_point(address)
        addresses = [point_to_string(point, address_format) for address_format in list(AddressFormat)]
        
        result = []
        for tx_hash, tx_data in self._pending_transactions.items():
            inputs_addresses = tx_data.get('inputs_addresses', [])
            if any(addr in addresses for addr in inputs_addresses):
                tx = await Transaction.from_hex(tx_data['tx_hex'], check_signatures)
                result.extend([{'tx_hash': tx_input.tx_hash, 'index': tx_input.index} for tx_input in tx.inputs])
        
        return result

    async def get_spendable_outputs(self, address: str, check_pending_txs: bool = False) -> List[TransactionInput]:
        point = string_to_point(address)
        addresses = [point_to_string(point, address_format) for address_format in list(AddressFormat)]
        
        result = []
        
        # Find unspent outputs for this address
        for tx_hash, index in self._unspent_outputs:
            if tx_hash in self._transactions:
                tx_data = self._transactions[tx_hash]
                outputs_addresses = tx_data.get('outputs_addresses', [])
                outputs_amounts = tx_data.get('outputs_amounts', [])
                
                if index < len(outputs_addresses) and outputs_addresses[index] in addresses:
                    # Check if not pending spent
                    if not check_pending_txs or (tx_hash, index) not in self._pending_spent_outputs:
                        amount = Decimal(outputs_amounts[index]) / SMALLEST
                        result.append(TransactionInput(tx_hash, index, amount=amount, public_key=point))
        
        return result

    async def get_address_balance(self, address: str, check_pending_txs: bool = False) -> Decimal:
        tx_inputs = await self.get_spendable_outputs(address, check_pending_txs=check_pending_txs)
        balance = sum([tx_input.amount for tx_input in tx_inputs], Decimal(0))
        
        if check_pending_txs:
            point = string_to_point(address)
            search_patterns = [point_to_bytes(string_to_point(address), address_format).hex() for address_format in list(AddressFormat)]
            addresses = [point_to_string(point, address_format) for address_format in list(AddressFormat)]
            
            for tx_hash, tx_data in self._pending_transactions.items():
                if any(pattern in tx_data['tx_hex'] for pattern in search_patterns):
                    tx = await Transaction.from_hex(tx_data['tx_hex'], check_signatures=False)
                    for i, tx_output in enumerate(tx.outputs):
                        if tx_output.address in addresses:
                            balance += tx_output.amount
        
        return balance

    async def get_address_spendable_outputs_delta(self, address: str, block_no: int) -> Tuple[List[TransactionInput], List[TransactionInput]]:
        point = string_to_point(address)
        addresses = [point_to_string(point, address_format) for address_format in list(AddressFormat)]
        
        unspent_outputs = []
        spending_txs = []
        
        # Find unspent outputs from blocks >= block_no
        for tx_hash, index in self._unspent_outputs:
            if tx_hash in self._transactions:
                tx_data = self._transactions[tx_hash]
                block_hash = tx_data.get('block_hash')
                if block_hash in self._blocks:
                    block_id = self._blocks[block_hash].get('id', 0)
                    if block_id >= block_no:
                        outputs_addresses = tx_data.get('outputs_addresses', [])
                        outputs_amounts = tx_data.get('outputs_amounts', [])
                        if index < len(outputs_addresses) and outputs_addresses[index] in addresses:
                            amount = Decimal(outputs_amounts[index]) / SMALLEST
                            unspent_outputs.append(TransactionInput(tx_hash, index, amount=amount, public_key=point))
        
        # Find spending transactions from blocks >= block_no
        for tx_hash, tx_data in self._transactions.items():
            block_hash = tx_data.get('block_hash')
            if block_hash in self._blocks:
                block_id = self._blocks[block_hash].get('id', 0)
                if block_id >= block_no:
                    inputs_addresses = tx_data.get('inputs_addresses', [])
                    if any(addr in addresses for addr in inputs_addresses):
                        tx = await Transaction.from_hex(tx_data['tx_hex'], False)
                        spending_txs.extend(tx.inputs)
        
        return unspent_outputs, spending_txs

    async def get_nice_transaction(self, tx_hash: str, address: str = None):
        print("Running get_nice_transaction")
        
        # Check if it's a confirmed transaction
        get_pending = False
        tx_data = None
        block_timestamp = None
        
        if tx_hash in self._transactions:
            tx_data = self._transactions[tx_hash]
            block_hash = tx_data.get('block_hash')
            if block_hash and block_hash in self._blocks:
                block_timestamp = self._blocks[block_hash].get('timestamp')
        elif tx_hash in self._pending_transactions:
            get_pending = True
            tx_data = self._pending_transactions[tx_hash]
        
        if tx_data is None:
            return None
        
        # Parse timestamps
        time_received = None
        if tx_data.get('time_received'):
            try:
                time_received_dt = datetime.fromisoformat(tx_data['time_received'].replace('Z', '+00:00'))
                time_received = int(round(time_received_dt.timestamp()))
            except:
                time_received = None
        
        time_confirmed = None
        if block_timestamp:
            try:
                if isinstance(block_timestamp, str):
                    time_confirmed_dt = datetime.fromisoformat(block_timestamp.replace('Z', '+00:00'))
                else:
                    time_confirmed_dt = block_timestamp
                time_confirmed = int(round(time_confirmed_dt.timestamp()))
            except:
                time_confirmed = None
        
        tx = await Transaction.from_hex(tx_data['tx_hex'], False)
        
        if isinstance(tx, CoinbaseTransaction):
            transaction = {
                'is_coinbase': True,
                'hash': tx_hash,
                'block_hash': tx_data.get('block_hash'),
                'time_mined': time_received,
            }
        else:
            delta = None
            if address is not None:
                public_key = string_to_point(address)
                delta = 0
                inputs_addresses = tx_data.get('inputs_addresses', [])
                for i, tx_input in enumerate(tx.inputs):
                    if i < len(inputs_addresses) and string_to_point(inputs_addresses[i]) == public_key:
                        print('getting related output for delta')
                        delta -= await tx_input.get_amount()
                for tx_output in tx.outputs:
                    if tx_output.public_key == public_key:
                        delta += tx_output.amount
            
            transaction = {
                'is_coinbase': False,
                'hash': tx_hash,
                'block_hash': tx_data.get('block_hash'),
                'message': tx.message.hex() if tx.message is not None else None,
                'time_received': time_received,
                'time_confirmed': time_confirmed,
                'inputs': [],
                'delta': delta,
                'fees': await tx.get_fees()
            }
            
            if get_pending:
                del transaction['time_confirmed']
            
            inputs_addresses = tx_data.get('inputs_addresses', [])
            for i, input_tx in enumerate(tx.inputs):
                input_address = inputs_addresses[i] if i < len(inputs_addresses) else None
                transaction['inputs'].append({
                    'index': input_tx.index,
                    'tx_hash': input_tx.tx_hash,
                    'address': input_address,
                    'amount': await input_tx.get_amount()
                })
        
        transaction['outputs'] = [{'address': output.address, 'amount': output.amount} for output in tx.outputs]
        return transaction


import hashlib
import json
import logging
import sys
from enum import Enum
from math import ceil
from datetime import datetime, timezone
from typing import Union

import base58
from fastecdsa.point import Point
from fastecdsa.util import mod_sqrt

from stellaris.constants import ENDIAN, CURVE

_print = print

#logging.basicConfig(level=logging.INFO if '--nologs' not in sys.argv else logging.WARNING)


def log(s):
    logging.getLogger('denaro').info(s)

def get_json(obj):
    return json.loads(
        json.dumps(obj, default=lambda o: getattr(o, 'as_dict', getattr(o, '__dict__', str(o))))
    )


def timestamp():
    return int(datetime.now(timezone.utc).replace(tzinfo=timezone.utc).timestamp())


def sha256(message: Union[str, bytes]):
    if isinstance(message, str):
        message = bytes.fromhex(message)
    return hashlib.sha256(message).hexdigest()


def byte_length(i: int):
    return ceil(i.bit_length() / 8.0)


def normalize_block(block) -> dict:
    block = dict(block)
    block['address'] = block['address'].strip(' ')
    # Convert timestamp string to datetime, set timezone, then get timestamp
    if isinstance(block['timestamp'], str):
        # If it's a string, parse it as a datetime first
        dt = datetime.fromisoformat(block['timestamp'].replace('Z', '+00:00'))
        block['timestamp'] = int(dt.replace(tzinfo=timezone.utc).timestamp())
    elif isinstance(block['timestamp'], datetime):
        # If it's already a datetime object
        block['timestamp'] = int(block['timestamp'].replace(tzinfo=timezone.utc).timestamp())
    else:
        # If it's already a number (timestamp), keep it as is
        block['timestamp'] = int(block['timestamp'])
    return block


def x_to_y(x: int, is_odd: bool = False):
    a, b, p = CURVE.a, CURVE.b, CURVE.p
    y2 = x ** 3 + a * x + b
    y_res, y_mod = mod_sqrt(y2, p)
    return y_res if y_res % 2 == is_odd else y_mod


class AddressFormat(Enum):
    FULL_HEX = 'hex'
    COMPRESSED = 'compressed'


def point_to_bytes(point: Point, address_format: AddressFormat = AddressFormat.FULL_HEX) -> bytes:
    if address_format is AddressFormat.FULL_HEX:
        return point.x.to_bytes(32, byteorder=ENDIAN) + point.y.to_bytes(32, byteorder=ENDIAN)
    elif address_format is AddressFormat.COMPRESSED:
        return string_to_bytes(point_to_string(point, AddressFormat.COMPRESSED))
    else:
        raise NotImplementedError()


def bytes_to_point(point_bytes: bytes) -> Point:
    if len(point_bytes) == 64:
        x, y = int.from_bytes(point_bytes[:32], ENDIAN), int.from_bytes(point_bytes[32:], ENDIAN)
        return Point(x, y, CURVE)
    elif len(point_bytes) == 33:
        specifier = point_bytes[0]
        x = int.from_bytes(point_bytes[1:], ENDIAN)
        return Point(x, x_to_y(x, specifier == 43), CURVE)
    else:
        raise NotImplementedError()


def bytes_to_string(point_bytes: bytes) -> str:
    point = bytes_to_point(point_bytes)
    if len(point_bytes) == 64:
        address_format = AddressFormat.FULL_HEX
    elif len(point_bytes) == 33:
        address_format = AddressFormat.COMPRESSED
    else:
        raise NotImplementedError()
    return point_to_string(point, address_format)


def point_to_string(point: Point, address_format: AddressFormat = AddressFormat.COMPRESSED) -> str:
    if address_format is AddressFormat.FULL_HEX:
        point_bytes = point_to_bytes(point)
        return point_bytes.hex()
    elif address_format is AddressFormat.COMPRESSED:
        x, y = point.x, point.y
        address = base58.b58encode((42 if y % 2 == 0 else 43).to_bytes(1, ENDIAN) + x.to_bytes(32, ENDIAN))
        return address if isinstance(address, str) else address.decode('utf-8')
    else:
        raise NotImplementedError()


def string_to_bytes(string: str) -> bytes:
    try:
        point_bytes = bytes.fromhex(string)
    except ValueError:
        point_bytes = base58.b58decode(string)
    return point_bytes


def string_to_point(string: str) -> Point:
    return bytes_to_point(string_to_bytes(string))


async def transaction_to_json(tx, verify: bool = False, address: str = None):
    from stellaris.transactions import CoinbaseTransaction
    if isinstance(tx, CoinbaseTransaction):
        transaction = {'is_coinbase': True, 'hash': tx.hash(), 'block_hash': tx.block_hash, 'outputs': []}
    else:
        delta = None
        if verify:
            await tx.verify(False)
            if address is not None:
                public_key = string_to_point(address)
                delta = 0
                for tx_input in tx.inputs:
                    if await tx_input.get_public_key() == public_key:
                        delta -= tx_input.amount
                for tx_output in tx.outputs:
                    if tx_output.public_key == public_key:
                        delta += tx_output.amount
        transaction = {'is_coinbase': False, 'hash': tx.hash(), 'block_hash': tx.block_hash, 'message': tx.message.hex() if tx.message is not None else None, 'inputs': [], 'outputs': [], 'delta': delta, 'fees': tx.fees}
        for input in tx.inputs:
            related_transaction = await transaction_to_json(await input.get_transaction()) if verify else None
            transaction['inputs'].append({
                'index': input.index,
                'tx_hash': input.tx_hash,
                'signature': input.get_signature() if input.signed is not None else None,
                'address': (await input.get_related_output()).address if verify else None,
                'amount': input.amount,
                'transaction': related_transaction
            })
    for output in tx.outputs:
        transaction['outputs'].append({
            'address': output.address,
            'amount': output.amount
        })
    return transaction



from decimal import Decimal
from io import BytesIO
from typing import List

from fastecdsa import keys
from stellaris.transactions import TransactionInput, TransactionOutput
from stellaris.constants import ENDIAN, SMALLEST, CURVE
from stellaris.utils.general import point_to_string, bytes_to_string, sha256

import struct

#print = ic


class Transaction:
    def __init__(self, inputs: List[TransactionInput], outputs: List[TransactionOutput], message: bytes = None, version: int = None):#, timestamp: int = None):
        if len(inputs) >= 256:
            raise Exception(f'You can spend max 255 inputs in a single transactions, not {len(inputs)}')
        if len(outputs) >= 256:
            raise Exception(f'You can have max 255 outputs in a single transactions, not {len(outputs)}')
        self.inputs = inputs
        self.outputs = outputs
        self.message = message
        if version is None:
            if all(len(tx_output.address_bytes) == 64 for tx_output in outputs):
                version = 1
            elif all(len(tx_output.address_bytes) == 33 for tx_output in outputs):
                version = 3
            else:
                raise NotImplementedError()
        if version > 3:
            raise NotImplementedError()
        self.version = version
        
        #self.timestamp = timestamp

        self._hex: str = None
        self.fees: Decimal = None
        self.tx_hash: str = None

    def hex(self, full: bool = True):
        inputs, outputs = self.inputs, self.outputs
        hex_inputs = ''.join(tx_input.tobytes().hex() for tx_input in inputs)
        hex_outputs = ''.join(tx_output.tobytes().hex() for tx_output in outputs)

        version = self.version

        self._hex = ''.join([
            version.to_bytes(1, ENDIAN).hex(),
            len(inputs).to_bytes(1, ENDIAN).hex(),
            hex_inputs,
            (len(outputs)).to_bytes(1, ENDIAN).hex(),
            hex_outputs
        ])

        if not full and (version <= 2 or self.message is None):
            return self._hex

        if self.message is not None:
            if version <= 2:
                self._hex += bytes([1, len(self.message)]).hex()
            else:
                self._hex += bytes([1]).hex()
                self._hex += (len(self.message)).to_bytes(2, ENDIAN).hex()
            self._hex += self.message.hex()
            if not full:
                return self._hex
        else:
            self._hex += (0).to_bytes(1, ENDIAN).hex()

        signatures = []
        for tx_input in inputs:
            signed = tx_input.get_signature()
            if signed not in signatures:
                signatures.append(signed)
                self._hex += signed

        return self._hex

    def hash(self):
        if self.tx_hash is None:
            self.tx_hash = sha256(self.hex())
        return self.tx_hash

    def _verify_double_spend_same_transaction(self):
        used_inputs = []
        for tx_input in self.inputs:
            input_hash = f"{tx_input.tx_hash}{tx_input.index}"
            if input_hash in used_inputs:
                return False
            used_inputs.append(input_hash)
        return True

    async def verify_double_spend(self):
        from stellaris.database import Database
        check_inputs = [(tx_input.tx_hash, tx_input.index) for tx_input in self.inputs]
        unspent_outputs = await Database.instance.get_unspent_outputs(check_inputs)
        return set(check_inputs) == set(unspent_outputs)

    async def verify_double_spend_pending(self):
        from stellaris.database import Database
        check_inputs = [(tx_input.tx_hash, tx_input.index) for tx_input in self.inputs]
        spent_outputs = await Database.instance.get_pending_spent_outputs(check_inputs)
        return spent_outputs == []

    async def _fill_transaction_inputs(self, txs=None) -> None:
        from stellaris.database import Database
        check_inputs = [tx_input.tx_hash for tx_input in self.inputs if tx_input.transaction is None and tx_input.transaction_info is None]
        if not check_inputs:
            return
        if txs is None:
            txs = await Database.instance.get_transactions_info(check_inputs)
        for tx_input in self.inputs:
            tx_hash = tx_input.tx_hash
            if tx_hash in txs:
                tx_input.transaction_info = txs[tx_hash]

    async def _check_signature(self):
        tx_hex = self.hex(False)
        checked_signatures = []
        for tx_input in self.inputs:
            if tx_input.signed is None:
                print('not signed')
                return False
            await tx_input.get_public_key()
            signature = (tx_input.public_key, tx_input.signed)
            if signature in checked_signatures:
                continue
            if not await tx_input.verify(tx_hex):
                print('signature not valid')
                return False
            checked_signatures.append(signature)
        return True

    def _verify_outputs(self):
        return (self.outputs or self.hash() == '915ddf143e14647ba1e04c44cf61e57084254c44cd4454318240f359a414065c') and all(tx_output.verify() for tx_output in self.outputs)

    async def verify(self, check_double_spend: bool = True) -> bool:
        if check_double_spend and not self._verify_double_spend_same_transaction():
            print('double spend inside same transaction')
            return False

        if check_double_spend and not await self.verify_double_spend():
            print('double spend')
            return False

        await self._fill_transaction_inputs()

        if not await self._check_signature():
            return False

        if not self._verify_outputs():
            print('invalid outputs')
            return False

        if await self.get_fees() < 0:
            print('We are not the Federal Reserve')
            return False

        return True

    async def verify_pending(self):
        return await self.verify() and await self.verify_double_spend_pending()

    def sign(self, private_keys: list = []):
        for private_key in private_keys:
            for input in self.inputs:
                if input.private_key is None and (input.public_key or input.transaction):
                    public_key = keys.get_public_key(private_key, CURVE)
                    input_public_key = input.public_key or input.transaction.outputs[input.index].public_key
                    if public_key == input_public_key:
                        input.private_key = private_key
        for input in self.inputs:
            if input.private_key is not None:
                input.sign(self.hex(False))
        return self

    async def get_fees(self):
        input_amount = 0
        for tx_input in self.inputs:
            input_amount += await tx_input.get_amount()

        output_amount = sum(tx_output.amount for tx_output in self.outputs)

        self.fees = input_amount - output_amount
        assert (self.fees * SMALLEST) % 1 == 0.0
        return self.fees
    
    # Timestamp handling might need to be changed in the future to address potential bugs and overflow issues.
    # Currently, the implementation "should" support Unix timestamps up to the maximum value of 4294967295.
    # This corresponds to the rollover date of 7 February 2106, 06:28:15 UTC, which is the upper limit
    # for dates representable by a 32-bit unsigned integer. 
    # Beyond this point, timestamp values will overflow...
    # Reference: https://en.wikipedia.org/wiki/Time_formatting_and_storage_bugs

    @staticmethod
    async def timestamp_to_bytes(timestamp):
        """
        Converts a Unix timestamp to a 4-byte representation.
    
        Args:
        - timestamp (int): The Unix timestamp to convert.
    
        Returns:
        - bytes: The 4-byte representation of the timestamp.
        """
        # Ensure the timestamp is within the 32-bit unsigned integer range.
        if timestamp < 0 or timestamp > 4294967295:
                                        
            raise ValueError("Timestamp must be between 0 and 4294967295.")
    
        return struct.pack('>I', timestamp)
    
    @staticmethod
    async def bytes_to_timestamp(byte_data):
        """
        Converts a 4-byte representation back into a Unix timestamp.
    
        Args:
        - byte_data (bytes): The 4 bytes to convert back to a timestamp.
    
        Returns:
        - int: The Unix timestamp.
        """
        if len(byte_data) != 4:
            raise ValueError("Input must be exactly 4 bytes.")

        return struct.unpack('>I', byte_data)[0]
    
    @staticmethod
    async def from_hex(hexstring: str, check_signatures: bool = True):#, set_timestamp:bool = False):

        #timestamp = None
        # Check if the timestamp should be extracted and converted from the hexstring.
        #if set_timestamp:
             # Extracts the last 8 characters as timestamp in hexadecimal, converts to bytes, then to actual timestamp.
        #    timestamp_hex = hexstring[-8:]
        #    timestamp_bytes = bytes.fromhex(timestamp_hex)
        #    timestamp = await Transaction.bytes_to_timestamp(timestamp_bytes)

            # Convert the rest of the hexstring (excluding the last 8 characters) to bytes
        #    tx_bytes = BytesIO(bytes.fromhex(hexstring[:-8]))
        #else:
            # If set_timestamp is False, convert the entire hexstring to bytes
        #    tx_bytes = BytesIO(bytes.fromhex(hexstring))

        tx_bytes = BytesIO(bytes.fromhex(hexstring))
        version = int.from_bytes(tx_bytes.read(1), ENDIAN)
        if version > 3:
            raise NotImplementedError()

        inputs_count = int.from_bytes(tx_bytes.read(1), ENDIAN)

        inputs = []

        for i in range(0, inputs_count):
            tx_hex = tx_bytes.read(32).hex()
            tx_index = int.from_bytes(tx_bytes.read(1), ENDIAN)
            inputs.append(TransactionInput(tx_hex, index=tx_index))

        outputs_count = int.from_bytes(tx_bytes.read(1), ENDIAN)

        outputs = []

        for i in range(0, outputs_count):
            pubkey = tx_bytes.read(64 if version == 1 else 33)
            amount_length = int.from_bytes(tx_bytes.read(1), ENDIAN)
            amount = int.from_bytes(tx_bytes.read(amount_length), ENDIAN) / Decimal(SMALLEST)
            outputs.append(TransactionOutput(bytes_to_string(pubkey), amount))

        specifier = int.from_bytes(tx_bytes.read(1), ENDIAN)
        if specifier == 36:
            assert len(inputs) == 1 and len(outputs) == 1
            # Import here to avoid circular import
            from stellaris.transactions.coinbase_transaction import CoinbaseTransaction
            return CoinbaseTransaction(inputs[0].tx_hash, outputs[0].address, outputs[0].amount)
        else:
            if specifier == 1:
                message_length = int.from_bytes(tx_bytes.read(1 if version <= 2 else 2), ENDIAN)
                message = tx_bytes.read(message_length)
            else:
                message = None
                assert specifier == 0

            signatures = []

            while True:
                signed = (int.from_bytes(tx_bytes.read(32), ENDIAN), int.from_bytes(tx_bytes.read(32), ENDIAN))
                if signed[0] == 0:
                    break
                signatures.append(signed)

            if len(signatures) == 1:
                for tx_input in inputs:
                    tx_input.signed = signatures[0]
            elif len(inputs) == len(signatures):
                for i, tx_input in enumerate(inputs):
                    tx_input.signed = signatures[i]
            else:
                if not check_signatures:
                    return Transaction(inputs, outputs, message, version)#, timestamp)
                index = {}
                for tx_input in inputs:
                    public_key = point_to_string(await tx_input.get_public_key())
                    if public_key not in index.keys():
                        index[public_key] = []
                    index[public_key].append(tx_input)

                for i, signed in enumerate(signatures):
                    for tx_input in index[list(index.keys())[i]]:
                        tx_input.signed = signed

            return Transaction(inputs, outputs, message, version)#, timestamp)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.hex() == other.hex()
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


import random
from asyncio import gather
from collections import deque
import os
from dotenv import dotenv_values
import re
import json
from decimal import Decimal
from datetime import datetime 

from asyncpg import UniqueViolationError
from fastapi import FastAPI, Body, Query
from fastapi.responses import RedirectResponse, Response

from httpx import TimeoutException
#from icecream import ic
from starlette.background import BackgroundTasks, BackgroundTask
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse

from slowapi import Limiter, _rate_limit_exceeded_handler
from slowapi.util import get_remote_address
from slowapi.errors import RateLimitExceeded

from stellaris.utils.general import timestamp, sha256, transaction_to_json
from stellaris.manager import create_block, get_difficulty, Manager, get_transactions_merkle_tree, \
    split_block_content, calculate_difficulty, clear_pending_transactions, block_to_bytes, get_transactions_merkle_tree_ordered
from stellaris.node.nodes_manager import NodesManager, NodeInterface
from stellaris.node.utils import ip_is_local
from stellaris.transactions import Transaction, CoinbaseTransaction
from stellaris.database import Database
from stellaris.constants import VERSION, ENDIAN


limiter = Limiter(key_func=get_remote_address)
app = FastAPI()
app.state.limiter = limiter
app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)
db: Database = None
NodesManager.init()
started = False
is_syncing = False
self_url = None

#print = ic

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_methods=["GET", "POST"],
    allow_headers=["*"],
)

config = dotenv_values(".env")

async def propagate(path: str, args: dict, ignore_url=None, nodes: list = None):
    global self_url
    self_node = NodeInterface(self_url or '')
    ignore_node = NodeInterface(ignore_url or '')
    aws = []
    for node_url in nodes or NodesManager.get_propagate_nodes():
        node_interface = NodeInterface(node_url)
        if node_interface.base_url == self_node.base_url or node_interface.base_url == ignore_node.base_url:
            continue
        aws.append(node_interface.request(path, args, self_node.url))
    for response in await gather(*aws, return_exceptions=True):
        print('node response: ', response)


async def create_blocks(blocks: list):
    _, last_block = await calculate_difficulty()
    last_block['id'] = last_block['id'] if last_block != {} else 0
    last_block['hash'] = last_block['hash'] if 'hash' in last_block else (30_06_2005).to_bytes(32, ENDIAN).hex()
    i = last_block['id'] + 1
    for block_info in blocks:
        block = block_info['block']
        txs_hex = block_info['transactions']
        txs = [await Transaction.from_hex(tx) for tx in txs_hex]
        #txs = [await Transaction.from_hex(tx, set_timestamp=True) for tx in txs_hex]
        for tx in txs:
            if isinstance(tx, CoinbaseTransaction):
                txs.remove(tx)
                break
        hex_txs = [tx.hex() for tx in txs]
        block['merkle_tree'] = get_transactions_merkle_tree(hex_txs) if i > 22500 else get_transactions_merkle_tree_ordered(hex_txs)
        block_content = block.get('content') or block_to_bytes(last_block['hash'], block)

        if i <= 22500 and sha256(block_content) != block['hash'] and i != 17972:
            from itertools import permutations
            for l in permutations(hex_txs):
                _hex_txs = list(l)
                block['merkle_tree'] = get_transactions_merkle_tree_ordered(_hex_txs)
                block_content = block_to_bytes(last_block['hash'], block)
                if sha256(block_content) == block['hash']:
                    break
        elif 131309 < i < 150000 and sha256(block_content) != block['hash']:
            for diff in range(0, 100):
                block['difficulty'] = diff / 10
                block_content = block_to_bytes(last_block['hash'], block)
                if sha256(block_content) == block['hash']:
                    break
        assert i == block['id']
        if not await create_block(block_content.hex() if isinstance(block_content, bytes) else block_content, txs, last_block):
            return False
        last_block = block
        i += 1
    return True


async def _sync_blockchain(node_url: str = None):
    print('sync blockchain')
    if not node_url:
        nodes = NodesManager.get_recent_nodes()
        if not nodes:
            return
        node_url = random.choice(nodes)
    node_url = node_url.strip('/')
    _, last_block = await calculate_difficulty()
    starting_from = i = await db.get_next_block_id()
    node_interface = NodeInterface(node_url)
    local_cache = None
    if last_block != {} and last_block['id'] > 500:
        remote_last_block = (await node_interface.get_block(i-1))['block']
        if remote_last_block['hash'] != last_block['hash']:
            print(remote_last_block['hash'])
            offset, limit = i - 500, 500
            remote_blocks = await node_interface.get_blocks(offset, limit)
            local_blocks = await db.get_blocks(offset, limit)
            local_blocks = local_blocks[:len(remote_blocks)]
            local_blocks.reverse()
            remote_blocks.reverse()
            print(len(remote_blocks), len(local_blocks))
            for n, local_block in enumerate(local_blocks):
                if local_block['block']['hash'] == remote_blocks[n]['block']['hash']:
                    print(local_block, remote_blocks[n])
                    last_common_block = local_block['block']['id']
                    local_cache = local_blocks[:n]
                    local_cache.reverse()
                    await db.remove_blocks(last_common_block + 1)
                    break

    #return
    limit = 1000
    while True:
        i = await db.get_next_block_id()
        try:
            blocks = await node_interface.get_blocks(i, limit)
        except Exception as e:
            print(e)
            #NodesManager.get_nodes().remove(node_url)
            NodesManager.sync()
            break
        try:
            _, last_block = await calculate_difficulty()
            if not blocks:
                print('syncing complete')
                if last_block['id'] > starting_from:
                    NodesManager.update_last_message(node_url)
                    if timestamp() - last_block['timestamp'] < 86400:
                        # if last block is from less than a day ago, propagate it
                        txs_hashes = await db.get_block_transaction_hashes(last_block['hash'])
                        await propagate('push_block', {'block_content': last_block['content'], 'txs': txs_hashes, 'block_no': last_block['id']}, node_url)
                break
            assert await create_blocks(blocks)
        except Exception as e:
            print(e)
            if local_cache is not None:
                print('sync failed, reverting back to previous chain')
                await db.delete_blocks(last_common_block)
                await create_blocks(local_cache)
            return


async def sync_blockchain(node_url: str = None):
    try:
        await _sync_blockchain(node_url)
    except Exception as e:
        print(e)
        return


@app.on_event("startup")
async def startup():
    global db
    global config
    db = await Database.create(
        user=config['STELLARIS_DATABASE_USER'] if 'STELLARIS_DATABASE_USER' in config else "stellaris" ,
        password=config['STELLARIS_DATABASE_PASSWORD'] if 'STELLARIS_DATABASE_PASSWORD' in config else 'stellaris',
        database=config['STELLARIS_DATABASE_NAME'] if 'STELLARIS_DATABASE_NAME' in config else "stellaris",
        host=config['STELLARIS_DATABASE_HOST'] if 'STELLARIS_DATABASE_HOST' in config else None
    )


@app.get("/")
async def root():
    return {"version": VERSION, "unspent_outputs_hash": await db.get_unspent_outputs_hash()}


async def propagate_old_transactions(propagate_txs):
    await db.update_pending_transactions_propagation_time([sha256(tx_hex) for tx_hex in propagate_txs])
    for tx_hex in propagate_txs:
        await propagate('push_tx', {'tx_hex': tx_hex})


@app.middleware("http")
async def middleware(request: Request, call_next):
    global started, self_url
    nodes = NodesManager.get_recent_nodes()
    hostname = request.base_url.hostname

    # Normalize the URL path by removing extra slashes
    normalized_path = re.sub('/+', '/', request.scope['path'])
    if normalized_path != request.scope['path']:
        url = request.url
        new_url = str(url).replace(request.scope['path'], normalized_path)
        #Redirect to normalized URL
        return RedirectResponse(new_url)

    if 'Sender-Node' in request.headers:
        NodesManager.add_node(request.headers['Sender-Node'])

    if nodes and not started or (ip_is_local(hostname) or hostname == 'localhost'):
        try:
            node_url = nodes[0]
            #requests.get(f'{node_url}/add_node', {'url': })
            j = await NodesManager.request(f'{node_url}/get_nodes')
            nodes.extend(j['result'])
            NodesManager.sync()
        except:
            pass

        if not (ip_is_local(hostname) or hostname == 'localhost'):
            started = True

            self_url = str(request.base_url).strip('/')
            try:
                nodes.remove(self_url)
            except ValueError:
                pass
            try:
                nodes.remove(self_url.replace("http://", "https://"))
            except ValueError:
                pass

            NodesManager.sync()

            try:
                await propagate('add_node', {'url': self_url})
                cousin_nodes = sum(await NodeInterface(url).get_nodes() for url in nodes)
                await propagate('add_node', {'url': self_url}, nodes=cousin_nodes)
            except:
                pass
    propagate_txs = await db.get_need_propagate_transactions()
    try:
        response = await call_next(request)
        response.headers['Access-Control-Allow-Origin'] = '*'
        if propagate_txs:
            response.background = BackgroundTask(propagate_old_transactions, propagate_txs)
        return response
    except:
        raise
        return {'ok': False, 'error': 'Internal error'}


@app.exception_handler(Exception)
async def exception_handler(request: Request, e: Exception):
    return JSONResponse(
        status_code=500,
        content={"ok": False, "error": f"Uncaught {type(e).__name__} exception"},
    )

transactions_cache = deque(maxlen=100)


@app.get("/push_tx")
@app.post("/push_tx")
async def push_tx(request: Request, background_tasks: BackgroundTasks, tx_hex: str = None, body=Body(False)):
    if body and tx_hex is None:
        tx_hex = body['tx_hex']
    tx = await Transaction.from_hex(tx_hex)
    if tx.hash() in transactions_cache:
        return {'ok': False, 'error': 'Transaction just added'}
    try:
        if await db.add_pending_transaction(tx):
            if 'Sender-Node' in request.headers:
                NodesManager.update_last_message(request.headers['Sender-Node'])
            background_tasks.add_task(propagate, 'push_tx', {'tx_hex': tx_hex})
            transactions_cache.append(tx.hash())
            return {'ok': True, 'result': 'Transaction has been accepted'}
        else:
            return {'ok': False, 'error': 'Transaction has not been added'}
    except UniqueViolationError:
        return {'ok': False, 'error': 'Transaction already present'}


@app.post("/push_block")
@app.get("/push_block")
async def push_block(request: Request, background_tasks: BackgroundTasks, block_content: str = '', txs='', block_no: int = None, body=Body(False)):
    if is_syncing:
        return {'ok': False, 'error': 'Node is already syncing'}
    if body:
        txs = body['txs']
        if 'block_content' in body:
            block_content = body['block_content']
        if 'id' in body:
            block_no = body['id']
        if 'block_no' in body:
            block_no = body['block_no']
    if isinstance(txs, str):
        txs = txs.split(',')
        if txs == ['']:
            txs = []
    previous_hash = split_block_content(block_content)[0]
    next_block_id = await db.get_next_block_id()
    if block_no is None:
        previous_block = await db.get_block(previous_hash)
        if previous_block is None:
            if 'Sender-Node' in request.headers:
                background_tasks.add_task(sync_blockchain, request.headers['Sender-Node'])
                return {'ok': False,
                        'error': 'Previous hash not found, had to sync according to sender node, block may have been accepted'}
            else:
                return {'ok': False, 'error': 'Previous hash not found'}
        block_no = previous_block['id'] + 1
    if next_block_id < block_no:
        background_tasks.add_task(sync_blockchain, request.headers['Sender-Node'] if 'Sender-Node' in request.headers else None)
        return {'ok': False, 'error': 'Blocks missing, had to sync according to sender node, block may have been accepted'}
    if next_block_id > block_no:
        return {'ok': False, 'error': 'Too old block'}
    final_transactions = []
    hashes = []
    for tx_hex in txs:
        if len(tx_hex) == 64:  # it's an hash
            hashes.append(tx_hex)
        else:
            final_transactions.append(await Transaction.from_hex(tx_hex))
    if hashes:
        pending_transactions = await db.get_pending_transactions_by_hash(hashes)
        if len(pending_transactions) < len(hashes):  # one or more tx not found
            if 'Sender-Node' in request.headers:
                background_tasks.add_task(sync_blockchain, request.headers['Sender-Node'])
                return {'ok': False,
                        'error': 'Transaction hash not found, had to sync according to sender node, block may have been accepted'}
            else:
                return {'ok': False, 'error': 'Transaction hash not found'}
        final_transactions.extend(pending_transactions)
    if not await create_block(block_content, final_transactions):
        return {'ok': False}

    if 'Sender-Node' in request.headers:
        NodesManager.update_last_message(request.headers['Sender-Node'])

    background_tasks.add_task(propagate, 'push_block', {
        'block_content': block_content,
        'txs': [tx.hex() for tx in final_transactions] if len(final_transactions) < 10 else txs,
        'block_no': block_no
    })
    return {'ok': True}


@app.get("/sync_blockchain")
@limiter.limit("10/minute")
async def sync(request: Request, node_url: str = None):
    global is_syncing
    if is_syncing:
        return {'ok': False, 'error': 'Node is already syncing'}
    is_syncing = True
    await sync_blockchain(node_url)
    is_syncing = False


async def sync_pending_transactions():
    """Sync pending transactions from other nodes if we have none"""
    if await db.get_pending_transactions_limit(1, hex_only=True):
        return  # We already have pending transactions
    
    nodes = NodesManager.get_recent_nodes()
    for node_url in nodes:
        try:
            node_interface = NodeInterface(node_url)
            response = await node_interface.request('get_pending_transactions', {})
            if response.get('ok') and response.get('result'):
                remote_txs = response['result'][:10]  # Get up to 10 transactions
                for tx_hex in remote_txs:
                    try:
                        tx = await Transaction.from_hex(tx_hex)
                        await db.add_pending_transaction(tx)
                        print(f"Synced transaction from {node_url}: {tx.hash()}")
                    except Exception as e:
                        print(f"Failed to sync transaction {tx_hex[:16]}...: {e}")
                if remote_txs:
                    print(f"Synced {len(remote_txs)} pending transactions from {node_url}")
                    break  # Stop after successfully syncing from one node
        except Exception as e:
            print(f"Failed to sync pending transactions from {node_url}: {e}")


LAST_PENDING_TRANSACTIONS_CLEAN = [0]
LAST_PENDING_SYNC = [0]


@app.get("/get_mining_info")
async def get_mining_info(background_tasks: BackgroundTasks, pretty: bool = False):
    Manager.difficulty = None
    difficulty, last_block = await get_difficulty()
    
    # Sync pending transactions if we have none and it's been a while
    if LAST_PENDING_SYNC[0] < timestamp() - 30:  # Sync every 30 seconds
        LAST_PENDING_SYNC[0] = timestamp()
        background_tasks.add_task(sync_pending_transactions)
    
    pending_transactions = await db.get_pending_transactions_limit(hex_only=True)
    pending_transactions = sorted(pending_transactions)
    if LAST_PENDING_TRANSACTIONS_CLEAN[0] < timestamp() - 600:
        print(LAST_PENDING_TRANSACTIONS_CLEAN[0])
        LAST_PENDING_TRANSACTIONS_CLEAN[0] = timestamp()
        background_tasks.add_task(clear_pending_transactions, pending_transactions)
    result = {'ok': True, 'result': {
        'difficulty': difficulty,
        'last_block': last_block,
        'pending_transactions': pending_transactions[:10],
        'pending_transactions_hashes': [sha256(tx) for tx in pending_transactions[:10]],
        'merkle_root': get_transactions_merkle_tree(pending_transactions[:10])
    }}
    return Response(content=json.dumps(result, indent=4, cls=CustomJSONEncoder), media_type="application/json") if pretty else result


@app.get("/get_address_info")
@limiter.limit("8/second")
async def get_address_info(request: Request, address: str, transactions_count_limit: int = Query(default=5, le=50), page: int = Query(default=1, ge=1), show_pending: bool = False, verify: bool = False, pretty: bool = False):    
    outputs = await db.get_spendable_outputs(address)
    balance = sum(output.amount for output in outputs)
    
     # Calculate offset for pagination
    offset = (page - 1) * transactions_count_limit
    
    # Fetch transactions with pagination
    transactions = await db.get_address_transactions(address, limit=transactions_count_limit, offset=offset, check_signatures=True) if transactions_count_limit > 0 else []

    result = {'ok': True, 'result': {
        'balance': "{:f}".format(balance),
        'spendable_outputs': [{'amount': "{:f}".format(output.amount), 'tx_hash': output.tx_hash, 'index': output.index} for output in outputs],
        'transactions': [await db.get_nice_transaction(tx.hash(), address if verify else None) for tx in transactions],
        #'transactions': [await db.get_nice_transaction(tx.hash(), address if verify else None) for tx in await db.get_address_transactions(address, limit=transactions_count_limit, check_signatures=True)] if transactions_count_limit > 0 else [],
        'pending_transactions': [await db.get_nice_transaction(tx.hash(), address if verify else None) for tx in await db.get_address_pending_transactions(address, True)] if show_pending else None,
        'pending_spent_outputs': await db.get_address_pending_spent_outputs(address) if show_pending else None
    }}
    return Response(content=json.dumps(result, indent=4, cls=CustomJSONEncoder), media_type="application/json") if pretty else result


@app.get("/add_node")
@limiter.limit("10/minute")
async def add_node(request: Request, url: str, background_tasks: BackgroundTasks):
    nodes = NodesManager.get_nodes()
    url = url.strip('/')
    if url == self_url:
        return {'ok': False, 'error': 'Recursively adding node'}
    if url in nodes:
        return {'ok': False, 'error': 'Node already present'}
    else:
        try:
            assert await NodesManager.is_node_working(url)
            background_tasks.add_task(propagate, 'add_node', {'url': url}, url)
            NodesManager.add_node(url)
            return {'ok': True, 'result': 'Node added'}
        except Exception as e:
            print(e)
            return {'ok': False, 'error': 'Could not add node'}


@app.get("/get_nodes")
async def get_nodes(pretty: bool = False):
    result = {'ok': True, 'result': NodesManager.get_recent_nodes()[:100]}
    return Response(content=json.dumps(result, indent=4, cls=CustomJSONEncoder), media_type="application/json") if pretty else result


@app.get("/get_pending_transactions")
async def get_pending_transactions(pretty: bool = False):
    result = {'ok': True, 'result': [tx.hex() for tx in await db.get_pending_transactions_limit(1000)]}
    return Response(content=json.dumps(result, indent=4, cls=CustomJSONEncoder), media_type="application/json") if pretty else result


@app.get("/get_transaction")
@limiter.limit("8/second")
async def get_transaction(request: Request, tx_hash: str, verify: bool = False, pretty: bool = False):
    tx = await db.get_nice_transaction(tx_hash)
    if tx is None:
        result = {'ok': False, 'error': 'Transaction not found'}
    else:
        result = {'ok': True, 'result': tx}
    return Response(content=json.dumps(result, indent=4, cls=CustomJSONEncoder), media_type="application/json") if pretty else result


@app.get("/get_block")
@limiter.limit("30/minute")
async def get_block(request: Request, block: str, full_transactions: bool = False, pretty: bool = False):
    if block.isdecimal():
        block_info = await db.get_block_by_id(int(block))
        if block_info is not None:
            block_hash = block_info['hash']
        else:
            result = {'ok': False, 'error': 'Block not found'}
    else:
        block_hash = block
        block_info = await db.get_block(block_hash)
    if block_info:
        result = {'ok': True, 'result': {
            'block': block_info,
            'transactions': await db.get_block_transactions(block_hash, hex_only=True) if not full_transactions else None,
            'full_transactions': await db.get_block_nice_transactions(block_hash) if full_transactions else None
        }}
    else:
        result = {'ok': False, 'error': 'Block not found'}
    return Response(content=json.dumps(result, indent=4, cls=CustomJSONEncoder), media_type="application/json") if pretty else result


@app.get("/get_blocks")
@limiter.limit("10/minute")
async def get_blocks(request: Request, offset: int, limit: int = Query(default=..., le=1000), pretty: bool = False):
    blocks = await db.get_blocks(offset, limit)
    result = {'ok': True, 'result': blocks}
    return Response(content=json.dumps(result, indent=4, cls=CustomJSONEncoder), media_type="application/json") if pretty else result

class CustomJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (Decimal, datetime)):
            return str(o)  # Convert types to string to prevent serialization errors
        return super().default(o)